from ehrapy.data._datasets import (
    breast_cancer_coimbra,
    breast_tissue,
    cervical_cancer_risk_factors,
    chronic_kidney_disease,
    dermatology,
    diabetes_130,
    echocardiogram,
    heart_disease,
    heart_failure,
    hepatitis,
    mimic_2,
    mimic_2_preprocessed,
    mimic_3_demo,
    parkinson_dataset_with_replicated_acoustic_features,
    parkinsons,
    parkinsons_disease_classification,
    parkinsons_telemonitoring,
    statlog_heart,
    synthea_1k_sample,
    thyroid,
)
